@@ -3,7 +3,7 @@ import pytest
 import unittest
 import pathlib
 
-from snap7.common import find_locally
+from snap7.common import _find_locally, load_library
 
 
 logging.basicConfig(level=logging.WARNING)
@@ -30,9 +30,13 @@ class TestCommon(unittest.TestCase):
         self.file.unlink()
 
     def test_find_locally(self):
-        file = find_locally(file_name_test.replace(".dll", ""))
+        file = _find_locally(file_name_test.replace(".dll", ""))
         self.assertEqual(file, str(self.BASE_DIR / file_name_test))
 
+    def test_raise_error_if_no_library(self):
+        with self.assertRaises(OSError):
+            load_library("wronglocation")
+
 
 if __name__ == "__main__":
     unittest.main()
